/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/game.c,v 1.15 90/09/28 12:21:03 rearl Exp $ */

/*
 * $Log:	game.c,v $
 * Revision 1.15  90/09/28  12:21:03  rearl
 * Added logging of interactive player commands.
 * 
 * Revision 1.14  90/09/16  04:42:11  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.13  90/09/15  22:20:04  rearl
 * Fixed non-GOD_PRIV problem.
 * 
 * Revision 1.12  90/09/13  06:22:31  rearl
 * Added optional argument to @dump for changing the dumpfile in emergencies.
 * 
 * Revision 1.11  90/09/10  02:19:54  rearl
 * Fixed a possible bug in database checkpointing, minimized
 * command-line space-smashing.
 * 
 * Revision 1.10  90/09/01  05:57:55  rearl
 * Fixed bugs in macro dumpfile.
 * 
 * Revision 1.9  90/08/27  03:24:52  rearl
 * Disk-based MUF source code, added environment support code.
 * 
 * Revision 1.8  90/08/11  04:00:20  rearl
 * *** empty log message ***
 * 
 * Revision 1.7  90/08/05  14:43:34  rearl
 * One more silly bug fixed in the command handler.
 * 
 * Revision 1.6  90/08/02  18:55:04  rearl
 * Fixed some calls to logging functions.
 * 
 * Revision 1.5  90/07/30  00:11:07  rearl
 * Added @owned command.
 * 
 * Revision 1.4  90/07/29  17:34:05  rearl
 * Fixed bug in = once and for all, also some bugs hiding in the command
 * processing switch.
 * 
 * Revision 1.3  90/07/21  16:32:23  casie
 * more log calls added
 * 
 * Revision 1.2  90/07/21  01:43:51  casie
 * added support for logging.
 * 
 * Revision 1.1  90/07/19  23:03:33  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <sys/wait.h>

#include "db.h"
#include "params.h"
#include "interface.h"
#include "match.h"
#include "externs.h"
#include "strings.h"

/* declarations */
static const char *dumpfile = 0;
static int epoch = 0;
static int alarm_triggered = 0;
static int alarm_block = 0;

static void fork_and_dump(void);
void dump_database(void);

void do_dump(dbref player, const char *newfile)
{
  char buf[BUFFER_LEN];
  
  if(Wizard(player)) {
    alarm_triggered = 1;
    if (*newfile
#ifdef GOD_PRIV
	&& God(player)
#endif /* GOD_PRIV */
	) {
      strcpy(dumpfile, newfile);
      sprintf(buf, "Dumping to file %s...", dumpfile);
    } else {
      sprintf(buf, "Dumping...");
    }
    notify(player, buf);
  } else {
    notify(player, "Sorry, you are in a no dumping zone.");
  }
}

void do_shutdown(dbref player)
{
  if(Wizard(player)) {
    log_status("SHUTDOWN: by %s\n", unparse_object(player, player));
    shutdown_flag = 1;
  } else {
    notify(player, "Your delusions of grandeur have been duly noted.");
    log_status("ILLEGAL SHUTDOWN: tried by %s\n", unparse_object(player, player));
  }
}

/* should be void, but it's defined as int */
static int alarm_handler(void)
{
  alarm_triggered = 1;
  if(!alarm_block) {
    fork_and_dump();
  }
  return 0;
}

static void dump_database_internal(void)
{
  char tmpfile[2048];
  FILE *f;
  
  sprintf(tmpfile, "%s.#%d#", dumpfile, epoch - 1);
  (void) unlink(tmpfile);		/* nuke our predecessor */
  
  sprintf(tmpfile, "%s.#%d#", dumpfile, epoch);
  
  if((f = fopen(tmpfile, "w")) != NULL) {
    db_write(f);
    fclose(f);
    if(rename(tmpfile, dumpfile) < 0) perror(tmpfile);
  } else {
    perror(tmpfile);
  }

  /* Write out the macros */

  sprintf(tmpfile, "%s.#%d#", MACRO_FILE, epoch - 1);
  (void) unlink(tmpfile);
  
  sprintf(tmpfile, "%s.#%d#", MACRO_FILE, epoch);
  
  if((f = fopen(tmpfile, "w")) != NULL) {
    macrodump(macrotop, f);
    fclose(f);
    if (rename(tmpfile, MACRO_FILE) < 0) perror(tmpfile);
  } else {
    perror(tmpfile);
  }
}

void panic(const char *message)
{
  char panicfile[2048];
  FILE *f;
  int i;
  
  log_status("PANIC: %s\n", message);
  fprintf(stderr, "PANIC: %s\n", message);
  
  /* turn off signals */
  for(i = 0; i < NSIG; i++) {
    signal(i, SIG_IGN);
  }
  
  /* shut down interface */
  emergency_shutdown();
  
  /* dump panic file */
  sprintf(panicfile, "%s.PANIC", dumpfile);
  if((f = fopen(panicfile, "w")) == NULL) {
    perror("CANNOT OPEN PANIC FILE, YOU LOSE");
#ifdef NOCOREDUMP
    _exit(135);
#else /* !NOCOREDUMP */
    signal(SIGIOT, SIG_DFL);
    abort();
#endif /* NOCOREDUMP */
  } else {
    log_status("DUMPING: %s\n", panicfile);
    fprintf(stderr, "DUMPING: %s\n", panicfile);
    db_write(f);
    fclose(f);
    log_status("DUMPING: %s (done)\n", panicfile);
    fprintf(stderr, "DUMPING: %s (done)\n", panicfile);
#ifdef NOCOREDUMP
    _exit(136);
#else /* !NOCOREDUMP */
    signal(SIGIOT, SIG_DFL);
    abort();
#endif /* NOCOREDUMP */
  }
}

void dump_database(void)
{
  epoch++;
  
  log_status("DUMPING: %s.#%d#\n", dumpfile, epoch);
  dump_database_internal();
  log_status("DUMPING: %s.#%d# (done)\n", dumpfile, epoch);
}

static void fork_and_dump(void)
{
  int child;
  
  epoch++;
  
  log_status("CHECKPOINTING: %s.#%d#\n", dumpfile, epoch);
#ifdef USE_VFORK
  child = vfork();
#else /* USE_VFORK */
  child = fork();
#endif /* USE_VFORK */
  if(child == 0) {
    /* in the child */
    close(0);		/* get that file descriptor back */
    dump_database_internal();
    _exit(0);		/* !!! */
  } else if(child < 0) {
    perror("fork_and_dump: fork()");
  }
  
  /* in the parent */
  /* reset alarm */
  alarm_triggered = 0;
  alarm(DUMP_INTERVAL);
}

static int reaper(void)
{
  union wait stat;
  
  while(wait3(&stat, WNOHANG, 0) > 0);
  return 0;
}

int init_game(const char *infile, const char *outfile)
{
  FILE *f;
  
  if ((f = fopen(MACRO_FILE, "r")) == NULL)
    log_status("INIT: Macro storage file %s is tweaked.\n", MACRO_FILE);
  else {
    macroload(f);
    fclose(f);
  }

  if((f = fopen(infile, "r")) == NULL) return -1;
  
  /* ok, read it in */
  log_status("LOADING: %s\n", infile);
  fprintf(stderr, "LOADING: %s\n", infile);
  if(db_read(f) < 0) return -1;
  log_status("LOADING: %s (done)\n", infile);
  fprintf(stderr, "LOADING: %s (done)\n", infile);
  
  /* everything ok */
  fclose(f);
  
  /* initialize random number generator */
  srandom(getpid());
  
  /* set up dumper */
  if(dumpfile) free((void *) dumpfile);
  dumpfile = alloc_string(outfile);
  signal(SIGALRM, alarm_handler);
  signal(SIGHUP, alarm_handler);
  signal(SIGCHLD, reaper);
  alarm_triggered = 0;
  alarm(DUMP_INTERVAL);
  
  return 0;
}

/* use this only in process_command */
#define Matched(string) { if(!string_prefix((string), command)) goto bad; }

void process_command(dbref player, char *command)
{
  char *arg1;
  char *arg2;
  char *full_command;
  char *p;			/* utility */
  char pbuf[BUFSIZ];
  char xbuf[BUFSIZ];
  
  if (command == 0) abort();
  
  /* robustify player */
  if(player < 0 || player >= db_top || Typeof(player) != TYPE_PLAYER) {
    log_status("process_command: bad player %d\n", player);
    return;
  }
  
#ifdef LOG_COMMANDS
  if (!(FLAGS(player) & INTERACTIVE) || DBFETCH(player)->sp.player.run)
    log_command("%s(%d) in %s(%d):%s %s\n", NAME(player), (int) player,
		NAME(DBFETCH(player)->location),
		(int) DBFETCH(player)->location,
		(FLAGS(player) & INTERACTIVE) ?
		" [interactive]" : " ", command);
#endif /* LOG_COMMANDS */
  
  /* block dump to prevent db inconsistencies from showing up */
  alarm_block = 1;
  
  if (FLAGS(player) & INTERACTIVE) {
      interactive(player, command);
      return;
    }
  
  /* eat leading whitespace */
  while(*command && isspace(*command)) command++;
  
  /* check for single-character commands */
  if(*command == SAY_TOKEN) {
    sprintf(pbuf,"say %s", command+1);
    command = &pbuf[0];
  } else if(*command == POSE_TOKEN) {
    sprintf(pbuf,"pose %s", command+1);
    command = &pbuf[0];
  }
  
  if(can_move(player, command)) {
    /* command is an exact match for an exit */
    do_move(player, command);
    *match_args = 0;
  } else {
    full_command = strcpy(xbuf, command);
    for (; *full_command && !isspace(*full_command); full_command++);
    if (*full_command) full_command++;
    
    /* find arg1 -- move over command word */
    for(arg1 = command; *arg1 && !isspace(*arg1); arg1++);
    /* truncate command */
    if(*arg1) *arg1++ = '\0';
    
    /* move over spaces */
    while(*arg1 && isspace(*arg1)) arg1++;
    
    /* find end of arg1, start of arg2 */
    for(arg2 = arg1; *arg2 && *arg2 != ARG_DELIMITER; arg2++);
    
    /* truncate arg1 */
    for(p = arg2 - 1; p >= arg1 && isspace(*p); p--) *p = '\0';
    
    /* go past delimiter if present */
    if(*arg2) *arg2++ = '\0';
    while(*arg2 && isspace(*arg2)) arg2++;
    
    switch(command[0]) {
    case '@':
      switch(command[1]) {
      case 'a':
      case 'A':
	/* @action, @attach */
	switch(command[2]) {
	case 'c':
	case 'C':
	  Matched("@action");
	  do_action(player, arg1, arg2);
	  break;
	case 't':
	case 'T':
	  Matched("@attach");
	  do_attach(player, arg1, arg2);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'b':
      case 'B':
	Matched("@boot");
	do_boot(player, arg1);
	break;
      case 'c':
      case 'C':
	/* chown, create */
	switch(command[2]) {
	case 'h':
	case 'H':
	  Matched("@chown");
	  do_chown(player, arg1, arg2);
	  break;
	case 'r':
	case 'R':
	  Matched("@create");
	  do_create(player, arg1, atoi(arg2));
	  break;
	default:
	  goto bad;
	}
	break;
      case 'd':
      case 'D':
	/* describe, dig, or dump */
	switch(command[2]) {
	case 'e':
	case 'E':
	  Matched("@describe");
	  do_describe(player, arg1, arg2);
	  break;
	case 'i':
	case 'I':
	  Matched("@dig");
	  do_dig(player, arg1, arg2);
	  break;
	case 'r':
	case 'R':
	  Matched("@drop");
	  do_drop_message(player, arg1, arg2);
	  break;
	case 'u':
	case 'U':
	  Matched("@dump");
	  do_dump(player, full_command);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'e':
      case 'E':
	Matched("@edit");
	do_edit(player, arg1);
	break;
      case 'f':
      case 'F':
	/* fail, find, or force */
	switch(command[2]) {
	case 'a':
	case 'A':
	  Matched("@fail");
	  do_fail(player, arg1, arg2);
	  break;
	case 'i':
	case 'I':
	  Matched("@find");
	  do_find(player, arg1);
	  break;
	case 'o':
	case 'O':
	  Matched("@force");
	  do_force(player, arg1, arg2);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'l':
      case 'L':
	/* lock or link */
	switch(command[2]) {
	case 'i':
	case 'I':
	  switch(command[3]) {
	  case 'n':
	  case 'N':
	    Matched("@link");
	    do_link(player, arg1, arg2);
	    break;
	  case 's':
	  case 'S':
	    Matched("@list");
	    match_and_list(player, arg1, arg2);
	    break;
	  default:
	    goto bad;
	  }
	  break;
	case 'o':
	case 'O':
	  Matched("@lock");
	  do_lock(player, arg1, arg2);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'n':
      case 'N':
	/* @name or @newpassword */
	switch(command[2]) {
	case 'a':
	case 'A':
	  Matched("@name");
	  do_name(player, arg1, arg2);
	  break;
	case 'e':
	case 'E':
	  if(strcmp(command, "@newpassword")) goto bad;
	  do_newpassword(player, arg1, arg2);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'o':
      case 'O':
	switch(command[2]) {
	case 'd':
	case 'D':
	  Matched("@odrop");
	  do_odrop(player, arg1, arg2);
	  break;
	case 'f':
	case 'F':
	  Matched("@ofail");
	  do_ofail(player, arg1, arg2);
	  break;
	case 'p':
	case 'P':
	  Matched("@open");
	  do_open(player, arg1, arg2);
	  break;
	case 's':
	case 'S':
	  Matched("@osuccess");
	  do_osuccess(player, arg1, arg2);
	  break;
	case 'w':
	case 'W':
	  Matched("@owned");
	  do_owned(player, full_command);
	  break;
	default:
	  goto bad;
	}
	break;
      case 'p':
      case 'P':
	switch(command[2])
	  {
	  case 'a':
	  case 'A':
	    Matched("@password");
	    do_password(player, arg1, arg2);
	    break;
#ifdef REGISTRATION
	  case 'c':
	  case 'C':
	    Matched("@pcreate");
	    do_pcreate(player, arg1, arg2);
	    break;
#endif /* REGISTRATION */
	  case 'r':
	  case 'R':
	    Matched("@prog");
	    do_prog(player, arg1);
	    break;
	  default:
	    goto bad;
	  }
	break;
#ifdef RECYCLE		
      case 'r':
      case 'R':
	Matched("@recycle");
	do_recycle(player, arg1);
	break;
#endif		
      case 's':
      case 'S':
	/* set, shutdown, success */
	switch(command[2]) {
	case 'e':
	case 'E':
	  Matched("@set");
	  do_set(player, arg1, arg2);
	  break;
	case 'h':
	case 'H':
	  if(strcmp(command, "@shutdown")) goto bad;
	  do_shutdown(player);
	  break;
	case 't':
	case 'T':
	  Matched("@stats");
	  do_stats(player, arg1);
	  break;
	case 'u':
	case 'U':
	  Matched("@success");
	  do_success(player, arg1, arg2);
	  break;
	default:
	  goto bad;
	}
	break;
      case 't':
      case 'T':
	switch(command[2]) {
	case 'e':
	case 'E':
	  Matched("@teleport");
	  do_teleport(player, arg1, arg2);
	  break;
	case 'o':
	case 'O':
	  if(strcmp(command, "@toad")) goto bad;
	  do_toad(player, arg1, arg2);
	  break;
	case 'r':
	case 'R':
	  Matched("@trace");
	  do_trace(player, arg1, atoi(arg2));
	  break;
	default:
	  goto bad;
	}
	break;
      case 'u':
      case 'U':
	if(string_prefix(command, "@unli")) {
	  do_unlink(player, arg1);
	} else if(string_prefix(command, "@unlo")) {
	  do_unlock(player, arg1);
	} else {
	  goto bad;
	}
	break;
      case 'w':
      case 'W':
	if(strcmp(command, "@wall")) goto bad;
	do_wall(player, full_command);
	break;
      default:
	goto bad;
      }
      break;
    case 'd':
    case 'D':
      Matched("drop");
      do_drop(player, arg1);
      break;
    case 'e':
    case 'E':
      Matched("examine");
      do_examine(player, arg1);
      break;
    case 'g':
    case 'G':
      /* get, give, go, or gripe */
      switch(command[1]) {
      case 'e':
      case 'E':
	Matched("get");
	do_get(player, arg1);
	break;
      case 'i':
      case 'I':
	Matched("give");
	do_give(player, arg1, atoi(arg2));
	break;
      case 'o':
      case 'O':
	Matched("goto");
	do_move(player, arg1);
	break;
      case 'r':
      case 'R':
	Matched("gripe");
	do_gripe(player, full_command);
	break;
      default:
	goto bad;
      }
      break;
    case 'h':
    case 'H':
      Matched("help");
      do_help(player);
      break;
    case 'i':
    case 'I':
      Matched("inventory");
      do_inventory(player);
      break;
    case 'k':
    case 'K':
      Matched("kill");
      do_kill(player, arg1, atoi(arg2));
      break;
    case 'l':
    case 'L':
      Matched("look");
      do_look_at(player, arg1);
      break;
    case 'm':
    case 'M':
      if (string_prefix(command, "move")) {
	do_move(player, arg1);
	break;
      } else {
	if (string_compare(command, "man"))
	  goto bad;
	do_man(player);
      }
      break;
    case 'n':
    case 'N':
      /* news */
      if(string_compare(command, "news")) goto bad;
      do_news(player);
      break;
    case 'p':
    case 'P':
      switch(command[1]) {
      case 'a':
      case 'A':
	Matched("page");
	do_page(player, arg1, arg2);
	break;
      case 'o':
      case 'O':
	Matched("pose");
	do_pose(player, full_command);
	break;
      case 'u':
      case 'U':
	Matched("put");
	do_drop(player, arg1);
	break;
      default:
	goto bad;
      }
      break;
    case 'r':
    case 'R':
      switch(command[1]) {
      case 'e':
      case 'E':
	Matched("read"); /* undocumented alias for look at */
	do_look_at(player, arg1);
	break;
      case 'o':
      case 'O':
	Matched("rob");
	do_rob(player, arg1);
	break;
      default:
	goto bad;
      }
      break;
    case 's':
    case 'S':
      /* say, "score" */
      switch(command[1]) {
      case 'a':
      case 'A':
	Matched("say");
	do_say(player, full_command);
	break;
      case 'c':
      case 'C':
	Matched("score");
	do_score(player);
	break;
      default:
	goto bad;
      }
      break;
    case 't':
    case 'T':
      switch(command[1]) {
      case 'a':
      case 'A':
	Matched("take");
	do_get(player, arg1);
	break;
      case 'h':
      case 'H':
	Matched("throw");
	do_drop(player, arg1);
	break;
      default:
	goto bad;
      }
      break;
    case 'w':
    case 'W':
      Matched("whisper");
      do_whisper(player, arg1, arg2);
      break;
    default:
    bad:
      notify(player, "Huh?  (Type \"help\" for help.)");
#ifdef LOG_FAILED_COMMANDS
      if(!controls(player, DBFETCH(player)->location)) {
	log_status("HUH from %s(%d) in %s(%d)[%s]: %s %s\n",
		   NAME(player), player, NAME(DBFETCH(player)->location),
		   DBFETCH(player)->location,
		   NAME(OWNER(DBFETCH(player)->location)), command,
		   full_command);
      }
#endif /* LOG_FAILED_COMMANDS */
      break;
    }
  }
  
  /* unblock alarms */
  alarm_block = 0;
  if(alarm_triggered) {
    fork_and_dump();
  }
}

#undef Matched
