/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/stringutil.c,v 1.10 90/09/28 12:25:08 rearl Exp $ */

/*
 * $Log:	stringutil.c,v $
 * Revision 1.10  90/09/28  12:25:08  rearl
 * Moved alloc_string() and alloc_prog_string() to here.
 * 
 * Revision 1.9  90/09/18  08:02:25  rearl
 * Speedups mainly, improved upper/lowercase lookup.
 * 
 * Revision 1.8  90/09/16  04:43:06  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.7  90/09/10  02:19:51  rearl
 * Introduced string compression of properties, for the
 * COMPRESS compiler option.
 * 
 * Revision 1.6  90/08/15  03:08:38  rearl
 * Messed around with pronoun_substitute.  Hopefully made it easier to use.
 * 
 * Revision 1.5  90/08/02  18:50:43  rearl
 * Fixed bug in capitalized self-substitutions.
 * 
 * Revision 1.4  90/08/02  02:17:16  rearl
 * Capital % substitutions now automatically capitalize the
 * corresponding self-sub property.  Example: %O -> Her.
 * 
 * Revision 1.3  90/07/29  17:45:08  rearl
 * Pronoun substitution for programs fixed.
 * 
 * Revision 1.2  90/07/22  04:28:33  casie
 * Added %r/%R substitutions for reflexive pronouns.
 * 
 * Revision 1.1  90/07/19  23:04:13  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"
#include "interface.h"

/* String utilities */

#include <ctype.h>
#include "externs.h"

extern const char *uppercase, *lowercase;
#define DOWNCASE(x) (lowercase[x])
#ifdef COMPRESS
extern const char *uncompress(const char *);
#endif /* COMPRESS */

int string_compare(register const char *s1, register const char *s2)
{
  while(*s1 && *s2 && DOWNCASE(*s1) == DOWNCASE(*s2)) s1++, s2++;
  
  return(DOWNCASE(*s1) - DOWNCASE(*s2));
}

int string_prefix(register const char *string, register const char *prefix)
{
  while(*string && *prefix && DOWNCASE(*string) == DOWNCASE(*prefix))
    string++, prefix++;
  return *prefix == '\0';
}

/* accepts only nonempty matches starting at the beginning of a word */
const char *string_match(register const char *src, register const char *sub)
{
  if(*sub != '\0') {
    while(*src) {
      if(string_prefix(src, sub)) return src;
      /* else scan to beginning of next word */
      while(*src && isalnum(*src)) src++;
      while(*src && !isalnum(*src)) src++;
    }
  }
  
  return 0;
}

/*
 * pronoun_substitute()
 *
 * %-type substitutions for pronouns
 *
 * %a/%A for absolute possessive (his/hers/its, His/Hers/Its)
 * %s/%S for subjective pronouns (he/she/it, He/She/It)
 * %o/%O for objective pronouns (him/her/it, Him/Her/It)
 * %p/%P for possessive pronouns (his/her/its, His/Her/Its)
 * %r/%R for reflexive pronouns (himself/herself/itself,
 *                                Himself/Herself/Itself)
 * %n    for the player's name.
 */
char *pronoun_substitute(dbref player, const char *str)
{
  char c;
  char prn[3];
  static char buf[BUFFER_LEN];
  char *result;
  const char *self_sub;               /* self substitution code */
  
  const static char *subjective[4] = { "", "it", "she", "he" };
  const static char *possessive[4] = { "", "its", "her", "his" };
  const static char *objective[4] = { "", "it", "her", "him" };
  const static char *reflexive[4] = { "", "itself", "herself", "himself" };
  const static char *absolute[4] = { "", "its", "hers", "his" };
  
  prn[0] = '%';
  prn[2] = '\0';
  
#ifdef COMPRESS
  str = uncompress(str);
#endif /* COMPRESS */

  result = buf;
  while (*str) {
    if(*str == '%') {
      *result = '\0';
      prn[1] = c = *(++str);
#ifdef COMPRESS
      self_sub = uncompress(get_property_class(player, prn));
#else /* !COMPRESS */
      self_sub = get_property_class(player, prn);
#endif /* COMPRESS */
      if (self_sub)
	{
	  strcat(result, self_sub);
	  if (isupper(prn[1]))
	    *result = toupper(*result);
	  result += strlen(result);
	  str++;
	}
      else if (genderof(player) == GENDER_UNASSIGNED) {
	switch(c) {
	case 'n':
	case 'N':
	case 'o':
	case 'O':
	case 's':
	case 'S':
	case 'r':
	case 'R':
	  strcat(result, NAME(player));
	  break;
	case 'a':
	case 'A':
	case 'p':
	case 'P':
	  strcat(result, NAME(player));
	  strcat(result, "'s");
	  break;
	default:
	  result[0] = *str;
	  result[1] = 0;
	  break;
	}
	str++;
	result += strlen(result);
      } else {
	switch (c) {
	case 'a':
	case 'A':
	  strcat(result, absolute[genderof(player)]);
	  break;
	case 's':
	case 'S':
	  strcat(result, subjective[genderof(player)]);
	  break;
	case 'p':
	case 'P':
	  strcat(result, possessive[genderof(player)]);
	  break;
	case 'o':
	case 'O':
	  strcat(result, objective[genderof(player)]);
	  break;
	case 'r':
	case 'R':
	  strcat(result, reflexive[genderof(player)]);
	  break;
	case 'n':
	case 'N':
	  strcat(result, NAME(player));
	  break;
	default:
	  *result = *str;
	  result[1] = '\0';
	  break;
	} 
	if(isupper(c) && islower(*result)) {
	  *result = toupper(*result);
	}
	
	result += strlen(result);
	str++;
      }
    } else {
      *result++ = *str++;
    }
  }
  *result = '\0';
  return buf;
} 

char *alloc_string(const char *string)
{
  char *s;
  
  /* NULL, "" -> NULL */
  if(string == 0 || *string == '\0') return 0;
  
  if((s = (char *) malloc(strlen(string)+1)) == 0) {
    abort();
  }
  strcpy(s, string);
  return s;
}

struct shared_string *
  alloc_prog_string(const char *s)
{
  struct shared_string *ss;
  int length;
  
  if(s == NULL || *s == '\0')
    return(NULL);
  
  length = strlen(s);
  if((ss = (struct shared_string *)
      malloc(sizeof(struct shared_string) + length)) == NULL)
    abort();
  ss->links = 1;
  ss->length = length;
  bcopy(s, ss->data, ss->length + 1);
  return(ss);
}
